/****************************************************************************
	
   Copyright (C) 1995 University of Utah

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation, 675 Massachusetts Avenue, Cambridge, MA 02139 USA;
   either version 1, or (at your option) any later version.

   The University of Utah disclaims all warranties with regard to this
   software, including all implied warranties of merchantability and fitness.
   In no event shall the University of Utah be liable for any special, indirect
   or consequential damages or any damages whatsoever resulting from loss of
   use, data or profits, whether in an action of contract, negligence or other
   tortious action, arising out of or in connection with the use or performance
   of this software.

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "control.h"



/* Initialize the Hamlet control database from the given file.
   The default is ~/.hamlet. */

Control::Control (char *fname) {


    /* Generate and save the filename. */

    if (fname == NULL) {
        char *home = getenv("HAMLETHOME");
	if (!home || !*home) home = getenv("HOME");
	if (!home) strcpy(filename, "");
	else {
	  strncpy(filename, home, MAXPATH);
	  strncat(filename, "/.hamlet", MAXPATH-strlen(filename));
	}
    }

    else
	strncpy(filename, fname, MAXPATH);


    /* In case the file read fails, we still need a valid count. */

    count = 0;


    /* Read the handle mappings. */

    FILE *file = fopen(filename, "r");
    if (!file) return;

    while (fscanf(file, "%s %ld %ld %d",
		  &(entries[count].handle), &(entries[count].window),
		  &(entries[count].time), &(entries[count].confirm)) == 4)
	if (++count >= MAXCACHE) break;

    fclose(file);

}



/* Save the control database to the file from whence it came. */

Control::~Control () {

    /* Open the backing file. */

    FILE *file = fopen(filename, "w");
    if (!file) return;


    /* Save the handle mappings. */

    for (int i = 0; i < count; i++)
	fprintf(file, "%s %ld %ld %d\n",
		entries[i].handle, entries[i].window,
		entries[i].time, entries[i].confirm);


    fclose(file);

}



/* Clear out the control database. */

void Control::reset () {
    count = 0;
}



/* Insert a new handle/window/confirmation mapping, or update an
   existing one. */

void Control::insert (char *handle, Window window, int confirm) {

    /* If there's not room, purge the oldest entry.  If there's already
       a handle, replace it.  Otherwise, add at the end. */

    int oldest = 0;
    int current = count;

    for (int i = 0; i < count; i++) {

	if (entries[i].time < entries[oldest].time)
	    oldest = i;

	if (strncmp(entries[i].handle, handle, MAXHANDLE) == 0) {
	    current = i;
	    break;
	}

    }

    if (current >= MAXCACHE)
	current = oldest;

    else if (current == count)
	count++;


    /* Add it in. */

    strncpy(entries[current].handle, handle, MAXHANDLE);
    entries[current].window = window;
    entries[current].time = time(NULL);
    entries[current].confirm = confirm;

};



/* Retrieve the window and confirmation associated with the
   handle.  Update the timestamp.  If window is None, there
   was no mapping. */

void Control::fetch (char *handle, Window &window, int &confirm) {

    window = None;

    for (int i = 0; i < count; i++)
	if (strcmp(entries[i].handle, handle) == 0) {
	    if (time(NULL) - entries[i].time > TIMEOUT) {
		entries[i] = entries[count-1];
		count--;
		return;
	    }
	    else {
		entries[i].time = time(NULL);
		window = entries[i].window;
		confirm = entries[i].confirm;
		return;
	    }
	}

}



/* Remove the mapping information associated with "handle". */

void Control::remove (char *handle) {

    for (int i = 0; i < count; i++)
	if (strcmp(handle, entries[i].handle) == 0) {
	    count--;
	    entries[i] = entries[count];
	    return;
	}

}
