/****************************************************************************

   Copyright (C) 1995 University of Utah

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation, 675 Massachusetts Avenue, Cambridge, MA 02139 USA;
   either version 1, or (at your option) any later version.

   The University of Utah disclaims all warranties with regard to this
   software, including all implied warranties of merchantability and fitness.
   In no event shall the University of Utah be liable for any special, indirect
   or consequential damages or any damages whatsoever resulting from loss of
   use, data or profits, whether in an action of contract, negligence or other
   tortious action, arising out of or in connection with the use or performance
   of this software.

*****************************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "comm.h"
#include "control.h"
#include "server.h"
#include "xdialog.h"


/* Gets the user's login id.  Returned value must be copied quickly. */

static char *getUserName () {
    struct passwd *pw;
    if ((pw = getpwuid(getuid())) == 0)
	return(getenv("USER"));
    else
	return(pw->pw_name);
}


/******************************************************************************
* int HashKey (char *key, char *user) {
* 
* 	This function returns a hash value for the given two strings.  Strings
* are hashed based on the numerical value of the characters.  The hash value
* is in the range 0-99999.
*
*	The hash function uses a polynomial random number generator.  Each
* character of the string causes the random number generator to cycle c times,
* where c is the value of the character.  For each character, one random
* number is added to a grand total.  The grand total is then mod'ed so that
* the key will be within the desired range (0-99999).
*
*	This is not the fastest of hash functions, but it is obscure and
* will produce drastically different results for very similar strings.
******************************************************************************/

static int HashKey (char *key, char *user) {

    long Seed = 0x387a9234;
    long Total = 0;
    int Count;
    char *theKey = new char[strlen(key) + strlen(user) + 1];
    strcpy(theKey, key);
    strcat(theKey, user);
    const char *Chars = theKey;

// Loop through the characters until the null terminator is reached.

  while (*Chars) {
  

// Cycle the polynomial random number generator c times.  Each cycle produces
//   8 new unexpected bits.

    Count = *Chars;
    
    while (Count--)
      Seed = (((Seed >> 3) ^ Seed) >> 20) | (Seed << 8);  
  
// Clip and add the current random number to the total.

    Total += Seed & 0x1ffff;
    
// Advance to the next character.

    Chars++;
  }
  
// Clip and return the total.

  delete theKey;

  return Total % 100000;

}



/* Reads and discards the characters in "file" up to and including the
   next newline or EOF.  Returns 1 if the end of file was encountered,
   zero otherwise. */

int eatLine (FILE *file) {
    while (getc(file) != '\n')
	if (feof(file)) return(1);
    return(0);
}



/* Assumes that the next line in "file" will be formatted as follows.  First
   will come the "prefix" string, followed by a white-space delimited string.

   If there is no line at all, returns -1.  If the line contains more than
   "length" characters, returns -2.

   Otherwise, reads the entire line into "line".  If the line begins
   with the prefix, it is ignored. If the line doesn't contain a
   space-delimited string following the (optional) prefix, returns 2.
   Otherwise, it returns 0 if there was a prefix and 1 otherwise.

   In any event, does the best it can at reading a space-delimited string
   into "command".

   The two strings "line" and "command" are assumed to have length "length". */

static int readLine (FILE *file, char *prefix, char *line, char *command,
		     int length) {

    /* Null everything out in case one of the reads fails. */ 

    line[0] = '\0';
    command[0] = '\0';

    if (!fgets(line, length, file)) return(-1);
    if ((strlen(line) == length-1) && line[strlen(line)-1] != '\n') return(-2);

    int count = sscanf(line, "%s", command);
    int result = 1;
    if (strcmp(command, prefix) == 0) {
      count = sscanf(line+strlen(prefix), "%s", command);
      result = 0;
    }

    if (count != 1) return(2);
    return(result);

}



/* Processes the Hamlet copytar directive encoded into "filename".  The
   first two lines of a hamlet copytar file are

   <!-- hamlet
   <!-- copytar

   The "<!--" part is optional.

   Beginning on the next line is a tar file. */

static void processCopytar (FILE *file, char *filename) {


    /* See if the user wants to have the tarfile copied. */

    xdialogtar xtar("Do you want to have the following\n"
		    "files copied into your home directory?",
		    filename,
		    3);
    int result = xtar.activate();


    /* If the answer is yes, strip out the leading hamlet stuff and
       pipe the remainder to tar.  Confirm what happened. */

    if (result == 1) {
	char *command = new char[strlen(filename) + 50];
	sprintf(command, "tail +%d %s | (cd; tar -xf - 2> /dev/null)",
		         3, filename);

	if (system(command))
	    notifyUser("Unable to copy the files");
    
	else
	    notifyUser("Files copied successfully");

	delete command;

    }

}



/* Processes the Hamlet script directive encoded into "file".  The
   first three lines of a hamlet script file are

   <!-- hamlet
   <!-- script
   <!-- full pathname of script

   When this function is called, the file is at the beginning of the
   third line.  (The "<!--" part is optional.)

   If there is a fourth line containing a lesson key and the script
   succeeds, create a hash value from the lesson key and the user and
   display it. */

static void processScript (FILE *file) {

    /* Make sure that the third line of the file consists of the
       HTML comment sequence, white space, and a single word pathname. */

    char line[201], pathname[201], key[201];
    int result = readLine(file, "<!--", line, pathname, 201);

    switch (result) {

    case -1:
	notifyUser("Hamlet file formatting error:\n",
		   "Contains only two lines");
	fclose(file);
	exit(1);

    case -2:
	notifyUser("Hamlet file formatting error:\n",
		   "End of third line not found within 200 characters");
	fclose(file);
	exit(1);

    case 2:
	notifyUser("Hamlet file formatting error:\n",
		   "Third line doesn't contain script name:\n\n",
		   line);
	fclose(file);
	exit(1);

    default:
	break;

    }

    /* See whether or not there is a fourth line. */

    result = readLine(file, "<!--", line, key, 201);

    if (result)	key[0] = '\0';

    fclose(file);


    /* Make sure that the pathname exists and is owned by user "hamlet". */

    struct stat buffer;
    if (stat(pathname, &buffer)) {
	notifyUser("Hamlet script file references non-existent script:\n\n",
		   pathname,
		   "\n");
	exit(1);
    }

    struct passwd *pass = getpwnam("hamlet");
    
    if (!pass || (buffer.st_uid != pass->pw_uid)) {
	notifyUser("Hamlet script file not owned by user hamlet:\n\n",
		   pathname,
		   "\n");
	exit(1);
    }


    /* Execute the script and pipe the output to a temporary file. */

    char command[200];
    sprintf(command, "%s > /tmp/hamlet.%d", pathname, getpid());
    result = system(command);
    sprintf(command, "/tmp/hamlet.%d", getpid());
    char banner[200];


    /* Do one thing or another depending on whether this is a keyscript
       file. */

    if (!key[0]) strcpy(banner, "");

    else if (result) strcpy(banner, "You did not pass the checkoff");

    else sprintf(banner,
		 "You passed the checkoff, and your key is %d\n",
		 HashKey(key, getUserName()));


    /* Display the results. */

    xdialogtext xtext(banner, command);
    xtext.activate();


    /* Delete the temporary file. */

    sprintf(command, "rm /tmp/hamlet.%d", getpid());
    system(command);
    
}
   


/* Processes the Hamlet reset directive. */

static void processReset (FILE *file) {
    Control control;
    control.reset();
    notifyUser("Hamlet has been reset\n");
    fclose(file);
}
    


/* Processes the Hamlet send directive encoded into "file" using "comm"
   as necessary.  The structure of a hamlet send file is

   <!-- hamlet
   <!-- send
   <!-- window
   <!-- characters to send go here
   more characters to send to here

   where the "<!-- " parts are optional.  Also, the "more characters"
   part is terminated by

   <HEAD>

   if the fourth line begins with the HTML comment sequence.
   
   When this function is called, the file is at the beginning of the
   third line. */

static void processSend (FILE *file, Communicator &comm) {


    /* Make sure that the third line of the file consists of the (optional)
       HTML comment sequence, white space, and a single word window name. */

    char line[201], window[201];
    int result = readLine(file, "<!--", line, window, 201);

    switch (result) {

    case -1:
	notifyUser("Hamlet file formatting error:\n"
		   "Contains only two lines\n");
	fclose(file);
	exit(1);

    case -2:
	notifyUser("Hamlet file formatting error:\n",
		   "End of third line not found within 200 characters");
	fclose(file);
	exit(1);

    case 2:
	notifyUser("Hamlet file formatting error:\n",
		   "Third line oesn't identify a window:\n\n",
		   line);
	fclose(file);
	exit(1);

    default:
	break;

    }


    /* Send all remaining lines.  The first line can optionally begin
       with the HTML comment sequence followed by a blank.  This is
       ignored.  If it does begin this way, terminate when we encounter
       a line consisting only of <BODY>.  This is done for purposes of
       backward compatiblity. */

    if (result == 0) {
      char comment[5];
      char space;
      int count = fscanf(file, "%4s%c", comment, &space);
      if ((count != 2) || strcmp(comment, "<!--") != 0 || (space != ' ')) {
	notifyUser("Hamlet file formatting error:\n",
		   "Command line doesn't begin with \"<!-- \"");
	fclose(file);
	exit(1);
      }
    }

    /* Transmit the command beginning from the current point in the file.
       Let it know whether to stop at <BODY> (result = 0) or not. */
    
    comm.send(window, file, result);

}




/* Processes the Hamlet directive encoded into "filename" using "comm"
   as necessary.  The beginning of a hamlet file is

   <HTML>
   <!-- command

   or

   <HTML>
   command

   or

   <!-- hamlet
   command

   or some such (the first line is ignored)
   where "command" is either "send", "copytar", or "script".  We use
   "command" to decide how to process the directive. */

void processButton (FILE *file, Communicator &comm, char *filename) {


    /* Ignore the first line. */

    if (eatLine(file)) {
	notifyUser("Hamlet file formatting error:\n",
		   "File is empty");
	fclose(file);
	exit(1);
    }


    /* Make sure that the second line of the file consists only of the
       HTML comment sequence and a single word command.  (The comment
       sequence is optional.) */

    char line[201], command[201];
    int result = readLine(file, "<!--", line, command, 201);

    switch (result) {

    case -1:
	notifyUser("Hamlet file formatting error:\n",
		   "File contains only one line");
	fclose(file);
	exit(1);

    case -2:
	notifyUser("Hamlet file formatting error:\n",
		   "End of second line not found within 200 characters");
	fclose(file);
	exit(1);

    case 2:
	notifyUser("Hamlet file formatting error:\n",
		   "Doesn't contain a valid command:\n\n",
		   line);
	fclose(file);
	exit(1);

    default:
	break;

    }
	

    /* Dispatch on the command name. */

    if (!strcmp(command, "send"))
	processSend(file, comm);

    else if (!strcmp(command, "copytar"))
	processCopytar(file, filename);

    else if (!strcmp(command, "script"))
	processScript(file);

    else if (!strcmp(command, "reset"))
	processReset(file);

    else {
	notifyUser("Hamlet file error:\n",
		   "Unknown Hamlet action \"",
		   command,
		   "\"");
	fclose(file);
	exit(1);
    }


}



/* Displays an error message with a popup window. */

void notifyUser (char *m1, char *m2, char *m3, char *m4) {

    char *message = new char[strlen(m1) + strlen(m2) + strlen(m3) +
                             strlen(m4) + 10];
    if (message) {
	strcpy(message, m1);
	if (m2) strcat(message, m2);
	if (m3) strcat(message, m3);
	if (m4) strcat(message, m4);
	xdialogok xd(message, "OK");
	xd.activate();
	delete message;
    }

}


/* Displays a message and a list of questions; returns the response. */

int askUser (char *questions[], char *m1, char *m2,
	      char *m3, char *m4, char *m5) {

    char *message = new char[strlen(m1) + strlen(m2) + strlen(m3) +
                             strlen(m4) + strlen(m5) + 10];
    int result = 1;

    if (message) {
	strcpy(message, m1);
	strcat(message, m2);
	strcat(message, m3);
	strcat(message, m4);
	strcat(message, m5);
	xdialogradio xr(message, questions);
	result = xr.activate();
	delete message;
    }

    return(result);

}
